'''Visualization functions'''


import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.patheffects as path_effects
import matplotlib.patches
import matplotlib.collections
from matplotlib import cm

import numpy as np

from .globalVariables import *

from . import (visibilityGraphAuxiliaryFunctions,
               clusteringFunctions,
               coreFunctions,
               extendedDataFrame,
               visibilityGraphCommunityDetection,
               utilityFunctions)

def saveFigure(fig, saveDir, label, extension, dpi, close=True):

    """Function primarily used internally to save and close figures

    Parameters:
        saveDir: str
            Path of directories to save the object to

        extension: str, Default '.png'
            Path of directories to save the object to
            
        dpi: int, Default 300
            Figure resolution if rasterized

        close: boolean: Default True
            Whether to close the figure after saving

    Returns:
        None

    Usage:
        saveFigure(fig, saveDir, label, extension, dpi)
    """

    utilityFunctions.createDirectories(saveDir)

    try:
        if not extension[0]=='.':
            extension = ''.join(['.', extension])
    except Exception as exception:
        print(exception)
        print('Chack figure extension/format')

    if extension in ['.png', '.jpeg', '.tiff']:

        fig.savefig(os.path.join(saveDir, label + extension), dpi=dpi)
    
    elif extension in ['.svg', '.eps', '.pdf']:

        fig.savefig(os.path.join(saveDir, label + extension))

    else:
        print('Unsupported format. Figure not saved')

    plt.close(fig)

    return None

def makeDataHistograms(df, saveDir, dataName, figsize=(8,8), range_min=np.min, range_max=np.max, includeTitle=True, title='Data @ timePoint:', fontsize=8, fontcolor='b', N_bins=100, color='b', extension='.png', dpi=300):

    """Make a histogram for each Series (time point) in a Dataframe.

    Parameters:
        df: pandas.DataFrame
            Data to visualize

        saveDir: str
            Path of directories to save the object to

        dataName: str
            Label to include in the file name

        figsize: tuple, Default (8,8)
            Size of the figure in inches

        range_min: str, Default 
            How to determine data minimum

        range_max: int, float or function, Default 
            How to determine data maximum

        includeTitle: boolean, Default True
            Path of directories to save the object to

        title: str, Default 'Data @ timePoint:'
            Text of the title

        fontsize: str, Default 8
            Fontsize of the labels
            
        fontcolor: str, Default 'b'
            Color of the title font
            
        N_bins: int, Default 100
            Number of bins in the histogram
            
        color: str, Default 'b'
            Color of the bars
            
        extension: str, Default '.png'
            Path of directories to save the object to
            
        dpi: int, Default 300
            Figure resolution if rasterized

    Returns:
        None

    Usage:
        makeDataHistograms(df, '/dir1', 'myData')
    """

    if not os.path.exists(saveDir):
        os.makedirs(saveDir)

    for timePoint in df.columns[:]:

        label = str(timePoint)

        subset = df[timePoint]
        subset = subset[~np.isnan(subset.values)].values

        if type(range_min) is types.FunctionType:
            range_min = range_min(subset)

        if type(range_max) is types.FunctionType:
            range_max = range_max(subset)

        hist_of_subset = scipy.stats.rv_histogram(np.histogram(subset, bins=N_bins, range=(range_min,range_max)))
        hist_data = hist_of_subset._hpdf / N_bins
        hist_bins = hist_of_subset._hbins

        fig, ax = plt.subplots(figsize=figsize)

        bar_bin_width = range_max / N_bins

        ax.bar(hist_bins, hist_data[:-1], width=0.9 * bar_bin_width, color=color, align='center')

        if includeTitle:
            ax.set_title(title + label, fontdict={'color': fontcolor})

        ax.set_xlabel('Values', fontsize=fontsize)
        ax.set_ylabel('Density', fontsize=fontsize)

        ax.set_xlim(range_min - 0.5 * bar_bin_width, range_max + 0.5 * bar_bin_width)

        fig.tight_layout()

        saveFigure(fig, saveDir, dataName + '_' + label + '_histogram', extension, dpi)

    return None

def makeLombScarglePeriodograms(df, saveDir, dataName, minNumberOfNonzeroPoints=5, oversamplingRate=100, figsize=(5,5), title1='TimeSeries Data', title2='Lomb-Scargle periodogram' , extension='.png', dpi=300):
        
    """Make a combined plot of the signal and its Lomb-Scargle periodogram
    for each pandas Series (time point) in a Dataframe.

    Parameters:
        df: pandas.DataFrame
            Data to visualize

        saveDir: str
            Path of directories to save the object to

        dataName: str
            Label to include in the file name

        minNumberOfNonzeroPoints: int, Default 5
            Minimum number of non-zero points in signal to use it
            
        oversamplingRate: int, Default 100
            Periodogram oversampling rate

        figsize: tuple, Default (8,8)
            Size of the figure in inches

        title1: str, Default 'TimeSeries Data'
            Text of the upper title

        title2: str, Default 'Lomb-Scargle periodogram'
            Text of the lower title

        extension: str, Default '.png'
            Path of directories to save the object to
            
        dpi: int, Default 300
            Figure resolution if rasterized

    Returns:
        None

    Usage:
        makeLombScarglePeriodograms(df, '/dir1', 'myData')
    """

    if not os.path.exists(saveDir):
        os.makedirs(saveDir)

    for geneIndex in range(len(df.index[:])):
        
        geneName = df.index[geneIndex].replace(':', '_')

        subset = df.iloc[geneIndex]
        subset = subset[subset > 0.]

        setTimes, setValues, inputSetTimes = subset.index.values, subset.values, df.columns.values

        if len(subset) < minNumberOfNonzeroPoints:

            print(geneName, ' skipped (only %s non-zero point%s), ' % (len(subset), 's' if len(subset) != 1 else ''), end=' ', flush=True)

            continue

        pgram = coreFunctions.LombScargle(setTimes, setValues, inputSetTimes, OversamplingRate=oversamplingRate)

        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize)

        ax1.plot(setTimes, setValues, 'bo', linewidth=3, markersize=5, markeredgecolor='k', markeredgewidth=2)

        zero_points = np.array(list(set(inputSetTimes) - set(setTimes)))
        ax1.plot(zero_points, np.zeros((len(zero_points),)), 'ro', linewidth=3, markersize=3, markeredgecolor='k', markeredgewidth=0)
        ax1.set_aspect('auto')

        minTime = np.min(inputSetTimes)
        maxTime = np.max(inputSetTimes)
        extraTime = (maxTime - minTime) / 10

        ax1.set_xlim(minTime - extraTime, maxTime + extraTime)
        ax1.set_title(title1)
    
        ax2.plot(2 * np.pi * pgram[0], pgram[1], 'r-', linewidth=1)
        ax2.set_aspect('auto')
        ax2.set_title(title2)

        fig.tight_layout()

        saveFigure(fig, saveDir, dataName + '_' + geneName + '_Lomb_Scargle_periodogram', extension, dpi)

    return None

def addVisibilityGraph(data, times, dataName='G1S1', coords=[0.05,0.95,0.05,0.95], 
                   numberOfVGs=1, groups_ac_colors=['b'], fig=None, numberOfCommunities=6, printCommunities=False, 
                   fontsize=None, nodesize=None, level=0.55, commLineWidth=0.5, lineWidth=1.0,
                   withLabel=True, withTitle=False, layout='circle', radius=0.07, noplot=False, horizontal=False, communities=None, minNumberOfCommunities=2, communitiesMethod='betweenness_centrality', direction='left', weight='distance'):
    
    """Draw a Visibility graph of data on a provided Matplotlib figure.
    We represent each timepoint in a series as a node.
    Temporal events are detected and indicated with solid blue 
    lines encompassing groups of points, or communities.
    The shortest path identifies nodes (i.e. timepoints) that display high 
    intensity, and thus dominate the global signal profile, are robust 
    to noise, and are likely drivers of the global temporal behavior.

    Parameters:
        data: 2d numpy.array
            Array of data to visualize

        times: 1d numpy.array
            Times corresponding to each data point, used for labels

        dataName: str, Default 'G1S1'
            label to include in file name

        coords: list, Default [0.05,0.95,0.05,0.95]
            Coordinates of location of the plot on the figure

        numberOfVGs: int, Default 1
            Number of plots to add to this figure

        groups_ac_colors: list, Default ['b']
            Colors corresponding to different groups of graphs

        fig: matplotlib.figure, Default None
            Figure object

        numberOfCommunities: int, Default 6
            Number of communities

        printCommunities: boolean, Default False
            Whether to print communities details to screen

        fontsize: float, Default None
            Size of labels

        nodesize: float, Default None
            Size of nodes

        level: float, Default 0.55
            Distance of the community lines to nodes

        commLineWidth: float, Default 0.5
            Width of the community lines

        lineWidth: float, Default 1.0
            Width of the edges between nodes

        withLabel: boolean, Default True
            Whether to include label on plot

        withTitle: boolean, Default False
            Whether to include title on plot

        layout: str, Default 'circle'
            Type of the layout. Other option is 'line'

        radius: float, Default 0.07
            Radius of the circle

        noplot: boolean, Default False
            Whether to make a plot or only calculate communities

        horizontal: boolean, Default False
            Whether to use horizontal or natural visibility graph. 

        communities: tuple, Default None
            A tuple containing communities sturcture of network, and networkx Graph:
                List of list, e.g. [[],[],...]

                networkx.Graph 

        minNumberOfCommunities: int, Default 2
            Number of communities to find depends on the number of splits.
            This parameter is ignored in methods that automatically
            estimate optimal number of communities.

        communitiesMethod: str, Default 'betweenness_centrality'
            String defining the method to use for communitiy detection:
                'Girvan_Newman': edge betweenness centrality based approach

                'betweenness_centrality': reflected graph node betweenness centrality based approach

                'WDPVG': weighted dual perspective visibility graph method (note to also set weight variable)

        direction:str, default 'left'
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.

                'left': nodes can only aggregate to the left side hubs, e.g. early hubs

                'right': nodes can only aggregate to the right side hubs, e.g. later hubs

        weight: str, Default 'distance'
            Type of weight for communitiesMethod='WDPVG':
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
            
    Returns:
        tuple
            (graph_nx, data, communities)

    Usage:
        addVisibilityGraph(exampleData, exampleTimes, fig=fig, fontsize=16, nodesize=700, 
                            level=0.85, commLineWidth=3.0, lineWidth=2.0, withLabel=False)
    """

    if len(data.shape)>1:
        data = extendedDataFrame.DataFrame(data=data).imputeMissingWithMedian().apply(lambda data: np.sum(data[data > 0.0]) / len(data), axis=0).values
    if communities is None:
        communities, graph_nx = clusteringFunctions.getCommunitiesOfTimeSeries(data, times, minNumberOfCommunities=minNumberOfCommunities, horizontal=horizontal, method=communitiesMethod, direction=direction, weight=weight) 
    else:
        communities, graph_nx = communities

    if printCommunities:
        print('Communities:')
        [print(community) for community in communities]
        print('\n')

    if noplot:
        return graph_nx, data, communities

    group = int(dataName[:dataName.find('S')].strip('G'))

    if fontsize is None:
        fontsize = 4. * (8. + 5.) / (numberOfVGs + 5.)
    
    if nodesize is None:
        nodesize = 30. * (8. + 5.) / (numberOfVGs + 5.)

    (x1,x2,y1,y2) = coords
    
    axisVG = fig.add_axes([x1,y1,x2 - x1,y2 - y1])
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list('GR', [(0, 1, 0), (1, 0, 0)], N=1000)

    if layout=='line':
        pos = {i:[float(i)/float(len(graph_nx)), 0.5] for i in range(len(graph_nx))}
    else:
        pos = nx.circular_layout(graph_nx)
        keys = np.array(list(pos.keys())[::-1])
        values = np.array(list(pos.values()))
        values = (values - np.min(values, axis=0))/(np.max(values, axis=0)-np.min(values, axis=0))
        keys = np.roll(keys, np.argmax(values.T[1]) - np.argmin(keys))
        pos = dict(zip(keys, values))

    keys = np.array(list(pos.keys()))
    values = np.array(list(pos.values()))

    shortest_path = nx.shortest_path(graph_nx, source=min(keys), target=max(keys))
    shortest_path_edges = [(shortest_path[i],shortest_path[i + 1]) for i in range(len(shortest_path) - 1)]

    if layout=='line':
        for edge in graph_nx.edges:
            l = np.array(pos[edge[0]])
            r = np.array(pos[edge[1]])

            if edge in shortest_path_edges:
                axisVG.add_artist(matplotlib.patches.Wedge((l+r)/2., 0.5*np.sqrt((l-r)[0]*(l-r)[0]+(l-r)[1]*(l-r)[1]), 0, 180, fill=False, edgecolor='y', linewidth=0.5*3.*lineWidth, alpha=0.7, width=0.001))

            axisVG.add_artist(matplotlib.patches.Wedge((l+r)/2., 0.5*np.sqrt((l-r)[0]*(l-r)[0]+(l-r)[1]*(l-r)[1]), 0, 180, fill=False, edgecolor='k', linewidth=0.5*lineWidth, alpha=0.7, width=0.001))

        nx.draw_networkx(graph_nx, pos=pos, ax=axisVG, node_color='y', edge_color='y', node_size=nodesize * 1.7, width=0., nodelist=shortest_path, edgelist=shortest_path_edges, with_labels=False)
        nx.draw_networkx(graph_nx, pos=pos, ax=axisVG, node_color=data, cmap=cmap, alpha=1.0, font_size=fontsize,  width=0., font_color='k', node_size=nodesize)
    else:
        nx.draw_networkx(graph_nx, pos=pos, ax=axisVG, node_color='y', edge_color='y', node_size=nodesize * 1.7, width=3.0*lineWidth, nodelist=shortest_path, edgelist=shortest_path_edges, with_labels=False)
        nx.draw_networkx(graph_nx, pos=pos, ax=axisVG, node_color=data, cmap=cmap, alpha=1.0, font_size=fontsize,  width=lineWidth, font_color='k', node_size=nodesize)

    if layout=='line':
        xmin, xmax = (-1.,1.)
        ymin, ymax = (-1.,1.)
    else:
        xmin, xmax = axisVG.get_xlim()
        ymin, ymax = axisVG.get_ylim()

    X, Y = np.meshgrid(np.arange(xmin, xmax, (xmax - xmin) / 300.), np.arange(ymin, ymax, (ymax - ymin) / 300.))

    def smooth(Z, N=7.):
        for ix in range(1,Z.shape[0]-1,1):
            Z[ix] = ((N-1.)*Z[ix] + (Z[ix-1] + Z[ix+1])/2.)/N
        return Z

    for icommunity, community in enumerate(communities):
        Z = np.exp(X ** 2 - Y ** 2) * 0.
        nX, nY = tuple(np.array([pos[node] for node in community]).T)
        for i in range(len(community)-1):
            p1, p2 = np.array([nX[i], nY[i]]), np.array([nX[i+1], nY[i+1]])

            for j in range(-2, 32):
                pm = p1 + (p2-p1)*float(j)/30.
                Z[np.where((X-pm[0])**2+(Y-pm[1])**2<=radius**2)] = 1.
        
        for _ in range(20):
            Z = smooth(smooth(Z).T).T

        CS = axisVG.contour(X, Y, Z, [level], linewidths=commLineWidth, alpha=0.8, colors=groups_ac_colors[group - 1])
        #axisVG.clabel(CS, inline=True,fontsize=4,colors=group_colors[group-1], fmt ={level:'C%s'%icommunity})

    if layout=='line':
        axisVG.set_xlim(-0.1,1.)
        axisVG.set_ylim(-0.1,1.)

    axisVG.spines['left'].set_visible(False)
    axisVG.spines['right'].set_visible(False)
    axisVG.spines['top'].set_visible(False)
    axisVG.spines['bottom'].set_visible(False)
    axisVG.set_xticklabels([])
    axisVG.set_yticklabels([])
    axisVG.set_xticks([])
    axisVG.set_yticks([])

    if withLabel:
        axisVG.text(axisVG.get_xlim()[1], (axisVG.get_ylim()[1] + axisVG.get_ylim()[0]) * 0.5, dataName, ha='left', va='center',
                    fontsize=8).set_path_effects([path_effects.Stroke(linewidth=0.4, foreground=groups_ac_colors[group - 1]),path_effects.Normal()])

    if withTitle:
        titleText = dataName + ' (size: ' + str(data.shape[0]) + ')' + ' min=%s max=%s' % (np.round(min(data),2), np.round(max(data),2))
        axisVG.set_title(titleText, fontsize=10)

    return graph_nx, data, communities

def makeVisibilityGraph(intensities, positions, saveDir, fileName, communities=None, minNumberOfCommunities=2, communitiesMethod='betweenness_centrality', direction='left', weight='distance', printCommunities=False,fontsize=16, nodesize=500, level=0.5, commLineWidth=3.0, lineWidth=2.0, layout='circle', horizontal=False, radius=0.03, figsize=(10,10), addColorbar=True, colorbarAxisCoordinates=[0.90,0.7,0.02,0.2], colorbarLabelsize=12, colorbarPrecision=2, extension='png', dpi=300):

    '''Make either horizontal or normal visibility graph of a time series using function addVisibilityGraph.
    We represent each timepoint in a series as a node.
    Temporal events are detected and indicated with solid blue 
    lines encompassing groups of points, or communities.
    The shortest path identifies nodes (i.e. timepoints) that display high 
    intensity, and thus dominate the global signal profile, are robust 
    to noise, and are likely drivers of the global temporal behavior.

    Parameters:
        intensities: 
            Data to plot
    
        positions: 
            Time points corresponding to data

        saveDir: str
            Path of directories to save the object to

        fileName: str
            Label to include in the file name

        horizontal: boolean, Default False
            Whether to use horizontal or natural visibility graph. Note that if communitiesMethod 'WDPVG' is set, this setting has no effect. 

        communities: tuple, Default None
            A tuple containing communities sturcture of network, and networkx Graph:
                List of list, e.g. [[],[],...]

                networkx.Graph 

        minNumberOfCommunities: int, Default 2
            Number of communities to find depends on the number of splits.
            This parameter is ignored in methods that automatically
            estimate optimal number of communities.

        communitiesMethod: str, Default 'betweenness_centrality'
            String defining the method to use for communitiy detection:
                'Girvan_Newman': edge betweenness centrality based approach

                'betweenness_centrality': reflected graph node betweenness centrality based approach

                'WDPVG': weighted dual perspective visibility graph method (note to also set weight variable)

        direction:str, default 'left'
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.

                'left': nodes can only aggregate to the left side hubs, e.g. early hubs

                'right': nodes can only aggregate to the right side hubs, e.g. later hubs


        weight: str, Default 'distance'
            Type of weight for communitiesMethod='WDPVG':
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5

        printCommunities: boolean, Default False
            Whether to print communities details to screen

        fontsize: float, Default 16
            Labels fontsize
                    
        nodesize: int, Default 500
            Node size
                    
        level: float, Default 0.5
            Level
                    
        commLineWidth: float, Default 3.0
            Communities lines width
                    
        lineWidth: float, Default 2.0
            Edge lines width
                    
        layout: str, Default 'circle'
            Type of layout, 'circle' or 'line'
                    
        horizontal: boolean, Default False
            Whether to make horizontal of normal visibility graph
                                
        radius: float, Default 0.03
            Rounding of the lines
                                
        figsize: tuple, Default (10,10)
            Figure size in inches
                                
        addColorbar: boolean, Default True
            Whether to add colorbar
                                
        colorbarAxisCoordinates: list, Default [0.90,0.7,0.02,0.2]
            colorbar axis coordinates
                                            
        colorbarLabelsize: float, Default 12
            Colorbar labels size
                                            
        colorbarPrecision: int, Default 2
            colorbar labels rounding
                                            
        extension: str, Default '.png'
            Figure extension
                                                        
        dpi: int, Default 300
            Figure resolution

    Returns:
        None

    Usage:
        makeVisibilityGraph(data, times, 'dir1/', 'myData')
    '''
    
    fig = plt.figure(figsize=figsize)

    addVisibilityGraph(intensities, positions, fig=fig, fontsize=fontsize, nodesize=nodesize, level=level, 
                        commLineWidth=commLineWidth, lineWidth=lineWidth, withLabel=False, layout=layout, 
                        printCommunities=printCommunities, radius=radius, horizontal=horizontal,communities=communities, minNumberOfCommunities=minNumberOfCommunities, communitiesMethod=communitiesMethod, direction=direction, weight=weight)

    addColorbarToFigure(fig, intensities, axisCoordinates=colorbarAxisCoordinates, labelsize=colorbarLabelsize, precision=colorbarPrecision)

    saveFigure(fig, saveDir, ('horizontal_' if horizontal else 'normal_') + fileName, extension, dpi)

    return

def makeVisibilityBarGraph(data, times, saveDir, fileName, AdjacencyMatrix=None, horizontal=False, barWidth=0.2, dotColor='b', barColor='r', arrowColor='k', id='', extension='.png', figsize=(8,4), dpi=300):

    """Bar-plot style visibility graph.
    Representing the intensities as bars, this is equivalent to connecting the top 
    of each bar to another top if there is a direct line-of-sight to that top. 
    The resulting visibility graph has characteristics that reflect the equivalent 
    time series temporal structure and can be used to identify trends.

    Parameters:
        data: 2d numpy.array
            Numpy array of floats

        times: 2d numpy.array
            Numpy array of floats

        fileName: str
            Path where to save the figure file

        fileName: str
            Name of the figure file to save

        AdjacencyMatrix: 2d numpy.array, Default None
            Adjacency matrix of network
        
        horizontal: boolean, default False
            Horizontal or normal visibility graph
            
        barWidth: float, default 0.2
            Horizontal or normal visibility graph
            
        dotColor: str, default 'b'
            Color of the data points
                        
        barColor: str, default 'r'
            Color of the bars
            
        arrowColor: str, default 'k'
            Color of lines

        id: str or int, default ''
            Label to add to the figure title
            
        extension: str, Default '.png'
            Figure format
            
        figsize: tuple of int, Default (8,4)
            Figure size in inches
            
        dpi: int, Default 300
            Figure resolution

    Returns:
        None

    Usage:
        makeVisibilityBarGraph(A, data, times, 'my_figure')
    """

    fig, ax = plt.subplots(figsize=figsize)
    
    if AdjacencyMatrix is None:
        if horizontal:
            A = visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfHVG(data)
        else:
            A = visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfNVG(data, times)
    else:
        A = AdjacencyMatrix

    ax.bar(times, data, width=barWidth, color=barColor, align='center', zorder=-np.inf)
    ax.scatter(times, data, color=dotColor)

    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            if i>j and A[i,j] == 1:
                if horizontal:
                    level1 = level2 = np.min([data[i],data[j]])
                else:
                    level1 = data[i]
                    level2 = data[j]

                ax.annotate(text='', xy=(times[i],level1), xytext=(times[j],level2), 
                            arrowprops=dict(arrowstyle='<->', shrinkA=0, shrinkB=0, linestyle='--'))

    ax.set_title('%s Time Series'%(id), fontdict={'color': arrowColor})
    ax.set_xlabel('Times', fontsize=8)
    ax.set_ylabel('Signal intensity', fontsize=8)
    ax.set_xticks(times)
    ax.set_xticklabels(times, fontsize=10, rotation=90)
    ax.set_yticks([])

    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(True)

    fig.tight_layout()

    saveFigure(fig, saveDir, ('horizontal_' if horizontal else 'normal_') + fileName, extension, dpi)

    return None

def makePlotOfPeak(data_all, scores, selected_peak, selected_peak_value, plotID):

    '''Plot peaks. Function used internally during certain data processing steps'''

    fig, ax = plt.subplots()

    ax.plot(data_all.T[0], data_all.T[1], 'g', lw=3)
    ax.plot(scores.T[0], scores.T[1], 'ro', ms=5)
    ax.plot(selected_peak, selected_peak_value, 'bo', alpha=0.5, ms=10)

    # saveFigure(fig, saveDir, 'spline_%s' % ('' if plotID == None else str(plotID)), extension, dpi)

    return

def addColorbarToFigure(fig, data, axisCoordinates=[0.90,0.7,0.02,0.2], cmap=None, norm=None, labelsize=12, precision=2):

    '''Add colorbar to figure
    
    Parameters:
        fig: matplotlib.figure
            Data to plot
                  
        cmap: matplotlib.colors.LinearSegmentedColormap, Default None
            Colormap to use
                                       
        norm: matplotlib.colors.Normalize, Default None
            Colormap normalization
                                
        axisCoordinates: list, Default [0.90,0.7,0.02,0.2]
            colorbar axis coordinates
                                            
        labelsize: float, Default 12
            Colorbar labels size
                                            
        precision: int, Default 2
            Colorbar labels rounding
                                            
    Returns:
        None

    Usage:
        addColorbarToFigure(fig, data)
    '''

    data = np.array(data)
    dataMin = np.min(data)
    dataMax = np.max(data)

    axisColor = fig.add_axes(axisCoordinates)

    if cmap is None:
        cmap=matplotlib.colors.LinearSegmentedColormap.from_list('GR', [(0, 1, 0), (1, 0, 0)], N=100)

    if norm is None:
        norm=matplotlib.colors.Normalize(vmin=dataMin, vmax=dataMax)

    mapp=cm.ScalarMappable(norm=norm, cmap=cmap)

    mapp.set_array(data)

    fig.colorbar(mapp, cax=axisColor, ticks=[dataMax,dataMin])

    axisColor.tick_params(labelsize=labelsize)

    axisColor.set_yticklabels([np.round(dataMax,precision), np.round(dataMin,precision)])
    
    return

def makeDendrogramHeatmapOfClusteringObject(ClusteringObject, saveDir, dataName, AutocorrNotPeriodogr=True, textScale=1.0, figsize=(12,8), extension='.png', dpi=300, xLabel='Time', plotLabel='Transformed Expression',horizontal=False, minNumberOfCommunities=2, communitiesMethod='WDPVG', direction='left', weight='distance'):

    """Make Dendrogram-Heatmap plot along with Visibility graphs.

    Parameters:
        ClusteringObject: 
            Clustering object

        saveDir: str
            Path of directories to save the object to

        dataName: str
            Label to include in the file name

        AutocorrNotPeriodogr: boolean, Default True
            Whether to use autocorrelation method instead of periodograms

        textScale: float, Default 1.0
            scaling of text size

        figsize: tuple, Default (12,8)
            Figure size in inches 
            
        extension: str, Default '.png'
            Figure format extension
            
        dpi: int, Default 300
            Figure resolution 
        
        xLabel: str, Default 'Time'
            Label for the x axis in the heatmap

        plotLabel: str, Default 'Transformed Expression'
            Label for the heatmap plot
        
        horizontal: boolean, Default False
            Whether to use horizontal or natural visibility graph. Note that if communitiesMethod 'WDPVG' is set, this setting has no effect. 

        minNumberOfCommunities: int, Default 2
            Number of communities to find depends on the number of splits.
            This parameter is ignored in methods that automatically
            estimate optimal number of communities.

        communitiesMethod: str, Default 'WDPVG'
            String defining the method to use for communitiy detection:
                'Girvan_Newman': edge betweenness centrality based approach

                'betweenness_centrality': reflected graph node betweenness centrality based approach

                'WDPVG': weighted dual perspective visibility graph method (note to also set weight variable)

        direction:str, default 'left'
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.

                'left': nodes can only aggregate to the left side hubs, e.g. early hubs

                'right': nodes can only aggregate to the right side hubs, e.g. later hubs

        weight: str, Default 'distance'
            Type of weight for communitiesMethod='WDPVG':
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5

    Returns:
        None

    Usage:
        makeDendrogramHeatmap(myObj, '/dir1', 'myData', AutocorrNotPeriodogr=True)
    """

    def addAutocorrelationDendrogramAndHeatmap(ClusteringObject, groupColors, fig, AutocorrNotPeriodogr=AutocorrNotPeriodogr):

        axisDendro = fig.add_axes([0.68,0.1,0.17,0.8], frame_on=False)

        n_clusters = len(ClusteringObject.keys()) - 1
        hierarchy.set_link_color_palette(groupColors[:n_clusters]) #gist_ncar #nipy_spectral #hsv
        origLineWidth = matplotlib.rcParams['lines.linewidth']
        matplotlib.rcParams['lines.linewidth'] = 0.5
        Y_ac = ClusteringObject['linkage']
        Z_ac = hierarchy.dendrogram(Y_ac, orientation='left',color_threshold=Y_ac[-n_clusters + 1][2])
        hierarchy.set_link_color_palette(None)
        matplotlib.rcParams['lines.linewidth'] = origLineWidth
        axisDendro.set_xticks([])
        axisDendro.set_yticks([])

        posA = ((axisDendro.get_xlim()[0] if n_clusters == 1 else Y_ac[-n_clusters + 1][2]) + Y_ac[-n_clusters][2]) / 2
        axisDendro.plot([posA, posA], [axisDendro.get_ylim()[0], axisDendro.get_ylim()[1]], 'k--', linewidth = 1)

        clusters = []
        order = []
        tempData = None
        for group in sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage']):
            for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
                subgroupData = ClusteringObject[group][subgroup]['dataAutocorr'].values
                tempData = subgroupData if tempData is None else np.vstack((tempData, subgroupData))
                clusters.extend([group for _ in range(subgroupData.shape[0])])
                order.extend(ClusteringObject[group][subgroup]['order'])

        tempData = tempData[np.argsort(order),:][Z_ac['leaves'],:].T[1:].T
        clusters = np.array(clusters)
        cluster_line_positions = np.where(clusters - np.roll(clusters,1) != 0)[0]

        for i in range(n_clusters - 1):
            posB = cluster_line_positions[i + 1] * (axisDendro.get_ylim()[1] / len(Z_ac['leaves'])) - 5. * 0
            axisDendro.plot([posA, axisDendro.get_xlim()[1]], [posB, posB], '--', color='black', linewidth = 1.0)
        
        axisDendro.plot([posA, axisDendro.get_xlim()[1]], [0., 0.], '--', color='k', linewidth = 1.0)
        axisDendro.plot([posA, axisDendro.get_xlim()[1]], [-0. + axisDendro.get_ylim()[1], -0. + axisDendro.get_ylim()[1]], '--', color='k', linewidth = 1.0)

        axisMatrixAC = fig.add_axes([0.78 + 0.07,0.1,0.18 - 0.075,0.8])
        imAC = axisMatrixAC.imshow(tempData, aspect='auto', vmin=np.min(tempData), vmax=np.max(tempData), origin='lower', cmap=plt.cm.bwr)
        for i in range(n_clusters - 1):
            axisMatrixAC.plot([-0.5, tempData.shape[1] - 0.5], [cluster_line_positions[i + 1] - 0.5, cluster_line_positions[i + 1] - 0.5], '--', color='black', linewidth = 1.0)
        axisMatrixAC.set_xticks([i for i in range(tempData.shape[1] - 1)])
        axisMatrixAC.set_xticklabels([i + 1 for i in range(tempData.shape[1] - 1)], fontsize=6*textScale)
        axisMatrixAC.set_yticks([])
        axisMatrixAC.set_xlabel('Lag' if AutocorrNotPeriodogr else 'Frequency', fontsize=axisMatrixAC.xaxis.label._fontproperties._size*textScale)
        axisMatrixAC.set_title('Autocorrelation' if AutocorrNotPeriodogr else 'Periodogram', fontsize=axisMatrixAC.title._fontproperties._size*textScale)

        axisColorAC = fig.add_axes([0.9 + 0.065,0.55,0.01,0.35])
        axisColorAC.tick_params(labelsize=6*textScale)
        plt.colorbar(imAC, cax=axisColorAC, ticks=[np.round(np.min(tempData),2),np.round(np.max(tempData),2)])

        return

    def addGroupDendrogramAndShowSubgroups(ClusteringObject, groupSize, bottom, top, group, groupColors, fig):

        Y = ClusteringObject[group]['linkage']

        n_clusters = len(sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']))
        print("Number of subgroups:", n_clusters)

        axisDendro = fig.add_axes([left, bottom, dx + 0.005, top - bottom], frame_on=False)

        hierarchy.set_link_color_palette([matplotlib.colors.rgb2hex(rgb[:3]) for rgb in cm.nipy_spectral(np.linspace(0, 0.5, n_clusters + 1))]) #gist_ncar #nipy_spectral #hsv
        origLineWidth = matplotlib.rcParams['lines.linewidth']
        matplotlib.rcParams['lines.linewidth'] = 0.5
        Z = hierarchy.dendrogram(Y, orientation='left',color_threshold=Y[-n_clusters + 1][2])
        hierarchy.set_link_color_palette(None)
        matplotlib.rcParams['lines.linewidth'] = origLineWidth
        axisDendro.set_xticks([])
        axisDendro.set_yticks([])

        posA = axisDendro.get_xlim()[0]/2 if n_clusters == groupSize else (Y[-n_clusters + 1][2] + Y[-n_clusters][2])/2 
        axisDendro.plot([posA, posA], [axisDendro.get_ylim()[0], axisDendro.get_ylim()[1]], 'k--', linewidth = 1)

        clusters = []
        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
            clusters.extend([subgroup for _ in range(ClusteringObject[group][subgroup]['data'].values.shape[0])])

        clusters = np.array(clusters)

        cluster_line_positions = np.where(clusters - np.roll(clusters,1) != 0)[0]

        for i in range(n_clusters - 1):
            posB = cluster_line_positions[i + 1] * (axisDendro.get_ylim()[1] / len(Z['leaves'])) - 5. * 0
            axisDendro.plot([posA, axisDendro.get_xlim()[1]], [posB, posB], '--', color='black', linewidth = 1.0)
        
        axisDendro.plot([posA, axisDendro.get_xlim()[1]], [0., 0.], '--', color='k', linewidth = 1.0)
        axisDendro.plot([posA, axisDendro.get_xlim()[1]], [-0. + axisDendro.get_ylim()[1], -0. + axisDendro.get_ylim()[1]], '--', color='k', linewidth = 1.0)

        axisDendro.text(axisDendro.get_xlim()[0], 0.5 * axisDendro.get_ylim()[1], 
                        'G%s:' % group + str(groupSize), fontsize=14*textScale).set_path_effects([path_effects.Stroke(linewidth=1, foreground=groupColors[group - 1]),path_effects.Normal()])

        return n_clusters, clusters, cluster_line_positions

    def addGroupHeatmapAndColorbar(data_loc, n_clusters, clusters, cluster_line_positions, bottom, top, group, groupColors, fig,xLabel = 'Time',plotLabel = 'Transformed Expression'):

        axisMatrix = fig.add_axes([left + 0.205, bottom, dx + 0.025 + 0.075, top - bottom])

        masked_array = np.ma.array(data_loc, mask=np.isnan(data_loc))

        cmap = matplotlib.colors.LinearSegmentedColormap.from_list('GR', [(0, 1, 0), (1, 0, 0)], N=1000) #plt.cm.prism #plt.cm.hsv_r #plt.cm.RdYlGn_r
        cmap.set_bad('grey')
        im = axisMatrix.imshow(masked_array, aspect='auto', origin='lower', vmin=np.min(data_loc[np.isnan(data_loc) == False]), vmax=np.max(data_loc[np.isnan(data_loc) == False]), cmap=cmap)

        for i in range(n_clusters - 1):
            posB = cluster_line_positions[i + 1]
            posBm = cluster_line_positions[i]
            axisMatrix.plot([-0.5, data_loc.shape[1] - 0.5], [posB - 0.5, posB - 0.5], '--', color='black', linewidth = 1.0)

        def add_label(pos, labelText):
            return axisMatrix.text(-1., pos, labelText, ha='right', va='center', fontsize=12.*textScale).set_path_effects([path_effects.Stroke(linewidth=0.4, foreground=groupColors[group - 1]),path_effects.Normal()])

        order = clusters[np.sort(np.unique(clusters,return_index=True)[1])] - 1


        for i in range(n_clusters - 1):
            if len(data_loc[clusters == i + 1]) >= 5.:
                try:
                    add_label((cluster_line_positions[np.where(order == i)[0][0]] + cluster_line_positions[np.where(order == i)[0][0] + 1]) * 0.5, 'G%sS%s:%s' % (group,i + 1,len(data_loc[clusters == i + 1])))
                except Exception as exception:
                    print(exception)
                    print('Label printing error!')
        if len(data_loc[clusters == n_clusters]) >= 5.:
            posC = axisMatrix.get_ylim()[0] if n_clusters == 1 else cluster_line_positions[n_clusters - 1]
            add_label((posC + axisMatrix.get_ylim()[1]) * 0.5, 'G%sS%s:%s' % (group,n_clusters,len(data_loc[clusters == n_clusters])))

        axisMatrix.set_xticks([])
        axisMatrix.set_yticks([])

        times = ClusteringObject[group][subgroup]['data'].columns.values

        if group == 1:
            axisMatrix.set_xticks(range(data_loc.shape[1]))
            axisMatrix.set_xticklabels([('' if (i%2==1 and textScale>1.3) else np.int(time)) for i, time in enumerate(np.round(times,1))], rotation=0, fontsize=6*textScale)
            axisMatrix.set_xlabel(xLabel, fontsize=axisMatrix.xaxis.label._fontproperties._size*textScale)

        if group == sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage'])[-1]:
            axisMatrix.set_title(plotLabel, fontsize=axisMatrix.title._fontproperties._size*textScale)

        axisColor = fig.add_axes([0.635 - 0.075 - 0.1 + 0.075,current_bottom + 0.01,0.01, max(0.01,(current_top - current_bottom) - 0.02)])
        plt.colorbar(im, cax=axisColor, ticks=[np.max(im._A),np.min(im._A)])
        axisColor.tick_params(labelsize=6*textScale)
        axisColor.set_yticklabels([np.round(np.max(im._A),2),np.round(np.min(im._A),2)])

        return

    signalsInClusteringObject = 0
    for group in sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage']):
        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
            signalsInClusteringObject += ClusteringObject[group][subgroup]['data'].shape[0]

    fig = plt.figure(figsize=(12,8))

    left = 0.02
    bottom = 0.1
    current_top = bottom
    dx = 0.2
    dy = 0.8

    groupColors = ['b','g','r','c','m','y','k']
    [groupColors.extend(groupColors) for _ in range(10)]

    addAutocorrelationDendrogramAndHeatmap(ClusteringObject, groupColors, fig)

    for group in sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage']):

        tempData = None
        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
            subgroupData = ClusteringObject[group][subgroup]['data'].values
            tempData = subgroupData if tempData is None else np.vstack((tempData, subgroupData))

        current_bottom = current_top
        current_top += dy * float(len(tempData)) / float(signalsInClusteringObject)

        if len(tempData)==1:
            n_clusters, clusters, cluster_line_positions = 1, np.array([]), np.array([])
        else:
            n_clusters, clusters, cluster_line_positions = addGroupDendrogramAndShowSubgroups(ClusteringObject, len(tempData), current_bottom, current_top, group, groupColors, fig)

        addGroupHeatmapAndColorbar(tempData, n_clusters, clusters, cluster_line_positions, current_bottom, current_top, group, groupColors, fig,xLabel=xLabel,plotLabel=plotLabel)

    data_list = []
    data_names_list = []
    for group in sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage']):
        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
            if ClusteringObject[group][subgroup]['data'].shape[0] >= 5:
                data_list.append(ClusteringObject[group][subgroup]['data'].values)
                data_names_list.append('G%sS%s' % (group, subgroup))

    times = ClusteringObject[group][subgroup]['data'].columns.values

    for indVG, (dataVG, dataNameVG) in enumerate(zip(data_list, data_names_list)):
        
        x_min = 0.57
        x_max = 0.66
        y_min = 0.1
        y_max = 0.9

        numberOfVGs = len(data_list)
        height = min((y_max - y_min) / numberOfVGs, (x_max - x_min) * (12. / 8.))
        x_displacement = (x_max - x_min - height / 1.5) * 0.5
        y_displacement = (y_max - y_min - numberOfVGs * height) / numberOfVGs

        coords = [x_min + x_displacement, x_min + x_displacement + height / (12. / 8.), y_min + indVG * height + (0.5 + indVG) * y_displacement, y_min + (indVG + 1) * height + (0.5 + indVG) * y_displacement]

        addVisibilityGraph(dataVG, times, dataNameVG, coords, numberOfVGs, groupColors, fig,horizontal=horizontal, minNumberOfCommunities=minNumberOfCommunities, communitiesMethod=communitiesMethod, direction=direction, weight=weight)
    
    saveFigure(fig, saveDir, dataName + '_DendrogramHeatmap', extension, dpi)

    return None
    
def plotHVGBarGraphDual(A, data, times, fileName, title='', fontsize=8, barwidth=0.05, figsize=(8,4), dpi=600):
    
    """Bar-plot style horizontal visibility graph with different link colors for different perspectives

    Parameters:
        A: 2d numpy.array
            Adjacency matrix
        
        data: 2d numpy.array
            Data used to make the visibility graph
        
        times: 1d numpy.array
            Times corresponding to data points
        
        fileName: str
            Name of the figure file to save
        
        title: str, Default ''
            Label to add to the figure title
        
        figsize: tuple of int, Default (8,4)        
            Figure size in inches      
            
        dpi: int, 600        
            Resolution of the image
            
        barwidth: float, Default 0.05
            The bar width

        fontsize:int, Default 8
            The text font size
        
    Returns:
        None

    Usage:
        PlotHorizontalVisibilityGraph(A, data, times, 'Figure.png', 'Test Data')
    """

    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(times, data, width = barwidth, color='k', align='center', zorder=-np.inf)
    ax.scatter(times, data, color='k')

    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            if i>j and A[i,j] > 0:
                if data[i] > 0 and data[j] >0:
                    level = np.min([data[i],data[j]])
                elif data[i] < 0 and data[j] < 0:
                    level = np.max([data[i],data[j]])
                else:
                    level = 0
                ax.annotate(text='', xy=(times[i],level), xytext=(times[j],level), 
                        arrowprops=dict(arrowstyle='->', shrinkA=0, shrinkB=0,linestyle='--', color='r'))
                            
            if i>j and A[i,j] < 0:
                if data[i] > 0 and data[j] >0:
                    level = np.min([data[i],data[j]])
                elif data[i] < 0 and data[j] < 0:
                    level = np.max([data[i],data[j]])
                else:
                    level = 0
                ax.annotate(text='', xy=(times[i],level), xytext=(times[j],level), 
                                arrowprops=dict(arrowstyle='->', shrinkA=0, shrinkB=0,linestyle='--', color='b'))
                    

    ax.set_title('%s'%(id), fontdict={'color': 'k'},fontsize=fontsize)
    ax.set_xlabel('Times', fontsize=fontsize)
    ax.set_ylabel('Signal intensity', fontsize=fontsize)
    ax.set_xticks(times)
    ax.set_xticklabels([str(item) for item in np.round(times,2)],fontsize=fontsize, rotation=0)
    ax.set_yticks([])
    ax.axhline(y=0, color='k')

    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(True)

    fig.tight_layout()
    fig.savefig(fileName, dpi=dpi)
    plt.close(fig)
    
    return None

def plotNVGBarGraphDual(A, data, times, fileName, title='', fontsize=8, barwidth=0.05, figsize=(8,4), dpi=600):

    """Bar-plot style natural visibility graph with different link colors for different perspectives
    
    Parameters:
        A: 2d numpy.array
            Adjacency matrix
        
        data: 2d numpy.array
            Data used to make the visibility graph
        
        times: 1d numpy.array
            Times corresponding to data points
        
        fileName: str
            Name of the figure file to save
        
        title: str, Default ''
            Label to add to the figure title
        
        figsize: tuple of int, Default (8,4)        
            Figure size in inches
                    
        dpi: int, 600        
            Resolution of the image

        barwidth: float, Default 0.05
            The bar width

        fontsize:int, Default 8
            The text font size
        
    Returns:
        None

    Usage:
        PlotVisibilityGraph(A, data, times, 'FIgure.png', 'Test Data')
    """

    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(times, data, width = barwidth, color='k', align='center', zorder=-np.inf)
    ax.scatter(times, data, color='k')

    for i in range(A.shape[0]):
        for j in range(A.shape[1]):
            if i>j and A[i,j] > 0:
                ax.annotate(text='', xy=(times[i],data[i]), xytext=(times[j],data[j]), 
                            arrowprops=dict(arrowstyle='->', shrinkA=0, shrinkB=0,linestyle='--',color='r'))
                            
            if i>j and A[i,j] < 0:
                ax.annotate(text='', xy=(times[i],data[i]), xytext=(times[j],data[j]), 
                            arrowprops=dict(arrowstyle='->', shrinkA=0, shrinkB=0,linestyle='--',color='b'))                

    ax.set_title('%s'%(title), fontdict={'color': 'k'},fontsize=fontsize)
    ax.set_xlabel('Times', fontsize=fontsize)
    ax.set_ylabel('Signal intensity', fontsize=fontsize)
    ax.set_xticks(times)
    ax.set_xticklabels([str(item) for item in np.round(times,2)],fontsize=fontsize, rotation=0)
    ax.set_yticks([])
    ax.axhline(y=0, color='k')

    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(True)

    fig.tight_layout()
    fig.savefig(fileName, dpi=dpi)
    plt.close(fig)

    return None
    



'''Clustering-related functions'''

import scipy.signal
import scipy.spatial.distance
from scipy.interpolate import UnivariateSpline

from sklearn.cluster import AgglomerativeClustering, KMeans, MiniBatchKMeans
import sklearn.metrics

from .globalVariables import *

from . import (coreFunctions,
               visibilityGraphAuxiliaryFunctions,
               visibilityGraphCommunityDetection,
               utilityFunctions)


def getEstimatedNumberOfClusters(data, cluster_num_min, cluster_num_max, trials_to_do, numberOfAvailableCPUs=4, plotID=None, printScores=False):

    """ Get estimated number of clusters using ARI with KMeans

    Parameters:
        data: 2d numpy.array
            Data to analyze

        cluster_num_min: int
            Minimum possible number of clusters

        cluster_num_max: int
            Maximum possible number of clusters

        trials_to_do: int
            Number of trials to do in ARI function

        numberOfAvailableCPUs: int, Default 4
            Number of processes to run in parallel

        plotID: str, Default None
            Label for the plot of peaks

        printScores: boolean, Default False
            Whether to print all scores

    Returns: 
        tuple
            Largest peak, other possible peaks.

    Usage:
        n_clusters = getEstimatedNumberOfClusters(data, 1, 20, 25)
    """

    def getPeakPosition(scores, makePlot=False, plotID=None):

        print()

        spline = UnivariateSpline(scores.T[0], scores.T[1])
        spline.set_smoothing_factor(0.005)
        xs = np.linspace(scores.T[0][0], scores.T[0][-1], 1000)
        data = np.vstack((xs, spline(xs))).T

        data_all = data.copy()
        data = data[data.T[0] > 4.]
        peaks = scipy.signal.find_peaks(data.T[1])[0]

        if len(peaks) == 0:
            selected_peak = 5
            print('WARNING: no peak found')
        else:
            selected_peak = np.round(data.T[0][peaks[np.argmax(data.T[1][peaks])]],0).astype(int)

        selected_peak_value = scores.T[1][np.argwhere(scores.T[0] == selected_peak)[0][0]]
        peaks = np.round(data.T[0][peaks],0).astype(int) if len(peaks) != 0 else peaks

        #if makePlot:
        #    makePlotOfPeak(data_all, scores, selected_peak, selected_peak_value, plotID)

        print(selected_peak, peaks)

        return selected_peak, peaks

    print('Testing data clustering in a range of %s-%s clusters' % (cluster_num_min,cluster_num_max))
                
    scores = utilityFunctions.runCPUs(numberOfAvailableCPUs, runForClusterNum, [(cluster_num, copy.deepcopy(data), trials_to_do) for cluster_num in range(cluster_num_min, cluster_num_max + 1)])

    if printScores: 
        print(scores)
                
    return getPeakPosition(scores, makePlot=True, plotID=plotID)[0]

def getNClustersFromLinkageElbow(Y):

    """ Get optimal number clusters from linkage.
    A point of the highest accelleration of the fusion coefficient of the given linkage.

    Parameters:
        Y: 2d numpy.array
            Linkage matrix

    Returns:
        int
            Optimal number of clusters

    Usage:
        n_clusters = getNClustersFromLinkageElbow(Y)
    """

    return np.diff(np.array([[nc, Y[-nc + 1][2]] for nc in range(2,min(50,len(Y)))]).T[1], 2).argmax() + 1 if len(Y) >= 5 else 1

def getNClustersFromLinkageSilhouette(Y, data, metric):

    """Determine the optimal number of cluster in data maximizing the Silhouette score.

    Parameters:
        Y: 2d numpy.array
            Linkage matrix

        data: 2d numpy.array
            Data to analyze

        metric: str or function
            Distance measure

    Returns:
        int
            Optimal number of clusters

    Usage:
        n_clusters = getNClustersFromLinkageSilhouette(Y, data, 'euclidean')
    """

    max_score = 0
    n_clusters = 1

    distmatrix = scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(data, metric=metric))

    for temp_n_clusters in range(2,10):
        print(temp_n_clusters, end='a, ', flush=True)
        temp_clusters = scipy.cluster.hierarchy.fcluster(Y, t=temp_n_clusters, criterion='maxclust')

        print(temp_n_clusters, end='b, ', flush=True)
        temp_score = sklearn.metrics.silhouette_score(distmatrix, temp_clusters, metric=metric)

        if temp_score>max_score:
            max_score = temp_score
            n_clusters = temp_n_clusters

    return n_clusters - 1

def runForClusterNum(arguments):
    
    """Calculate Adjusted Rand Index of the data for a range of cluster numbers.

    Parameters:
        arguments: tuple
            A tuple of three parameters int the form (cluster_num, data_array, trials_to_do), where
                cluster_num: int
                    Maximum number of clusters

                data_array: 2d numpy.array
                    Data to test

                trials_to_do: int
                    Number of trials for each cluster number

    Returns:
        1d numpy.array
            Numpy array

    Usage:
        instPool = multiprocessing.Pool(processes = NumberOfAvailableCPUs)

        scores = instPool.map(runForClusterNum, [(cluster_num, copy.deepcopy(data), trials_to_do) for cluster_num in range(cluster_num_min, cluster_num_max + 1)])

        instPool.close()

        instPool.join()
    """

    np.random.seed()

    cluster_num, data_array, trials_to_do = arguments

    print(cluster_num, end=', ', flush=True)

    labels = [KMeans(n_clusters=cluster_num).fit(data_array).labels_ for i in range(trials_to_do)]

    agreement_matrix = np.zeros((trials_to_do,trials_to_do))

    for i in range(trials_to_do):
        for j in range(trials_to_do):
            agreement_matrix[i, j] = sklearn.metrics.adjusted_rand_score(labels[i], labels[j]) if agreement_matrix[j, i] == 0 else agreement_matrix[j, i]

    selected_data = agreement_matrix[np.triu_indices(agreement_matrix.shape[0],1)]

    return np.array((cluster_num, np.mean(selected_data), np.std(selected_data)))

def getGroupingIndex(data, n_groups=None, method='weighted', metric='correlation', significance='Elbow'):

    """Cluster data into N groups, if N is provided, else determine N
    return: linkage matrix, cluster labels, possible cluster labels.

    Parameters:
        data: 2d numpy.array
            Data to analyze

        n_groups: int, Default None
            Number of groups to split data into

        method: str, Default 'weighted'
            Linkage calculation method

        metric: str, Default 'correlation'
            Distance measure

        significance: str, Default 'Elbow'
            Method for determining optimal number of groups and subgroups

    Returns:
        tuple
            Linkage matrix, cluster index, possible groups

    Usage:
        x, y, z = getGroupingIndex(data, method='weighted', metric='correlation', significance='Elbow')
    """

    Y = hierarchy.linkage(data, method=method, metric=metric, optimal_ordering=False)

    if n_groups == None:
        if significance=='Elbow':
            n_groups = getNClustersFromLinkageElbow(Y)
        elif significance=='Silhouette':
            n_groups = getNClustersFromLinkageSilhouette(Y, data, metric)

    print('n_groups:', n_groups)

    labelsClusterIndex = scipy.cluster.hierarchy.fcluster(Y, t=n_groups, criterion='maxclust')

    groups = np.sort(np.unique(labelsClusterIndex))

    print([np.sum(labelsClusterIndex == group) for group in groups])

    return Y, labelsClusterIndex, groups

def makeClusteringObject(df_data, df_data_autocorr, method='weighted', metric='correlation', significance='Elbow'):

    """Make a clustering Groups-Subgroups dictionary object.

    Parameters:
        df_data: pandas.DataFrame
            Data to analyze in DataFrame format

        df_data_autocorr: pandas.DataFrame
            Autocorrelations or periodograms in DataFrame format

        method: str, Default 'weighted'
            Linkage calculation method

        metric: str, Default 'correlation'
            Distance measure

        significance: str, Default 'Elbow'
            Method for determining optimal number of groups and subgroups

    Returns:
        dictionary
            Clustering object

    Usage:
        myObj = makeClusteringObject(df_data, df_data_autocorr, significance='Elbow')
    """

    def getSubgroups(df_data, metric, significance):

        Y = hierarchy.linkage(df_data.values, method=method, metric=metric, optimal_ordering=True)
        leaves = hierarchy.dendrogram(Y, no_plot=True)['leaves']

        if significance=='Elbow':
            n_clusters = getNClustersFromLinkageElbow(Y)
        elif significance=='Silhouette':
            n_clusters = getNClustersFromLinkageSilhouette(Y, df_data.values, metric)

        print('n_subgroups:', n_clusters)

        clusters = scipy.cluster.hierarchy.fcluster(Y, t=n_clusters, criterion='maxclust')[leaves]

        return {cluster:df_data.index[leaves][clusters==cluster] for cluster in np.unique(clusters)}, Y

    ClusteringObject = {}

    try:
        grouping = getGroupingIndex(df_data_autocorr.values, method=method, metric=metric, significance=significance)
    except Exception as exception:
        print(exception)
        print('Returning None')
        return None
        
    ClusteringObject['linkage'], labelsClusterIndex, groups = grouping

    for group in groups:
        signals = df_data.index[labelsClusterIndex==group]

        ClusteringObject[group], ClusteringObject[group]['linkage'] = ({1: signals}, None) if len(signals)==1 else getSubgroups(df_data.loc[signals], coreFunctions.metricCommonEuclidean, significance)

        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):
            ClusteringObject[group][subgroup] = {'order':[np.where([temp==signal for temp in df_data.index.values])[0][0] for signal in list(ClusteringObject[group][subgroup])],
                                                 'data':df_data.loc[ClusteringObject[group][subgroup]], 
                                                 'dataAutocorr':df_data_autocorr.loc[ClusteringObject[group][subgroup]]}

    return ClusteringObject

def exportClusteringObject(ClusteringObject, saveDir, dataName, includeData=True, includeAutocorr=True):

    """Export a clustering Groups-Subgroups dictionary object to a SpreadSheet.
    Linkage data is not exported.

    Parameters:
        ClusteringObject: dictionary
            Clustering object

        saveDir: str
            Path of directories to save the object to

        dataName: str
            Label to include in the file name

        includeData: boolean, Default True
            Export data

        includeAutocorr: boolean, Default True
            Export autocorrelations of data

    Returns:
        str
            File name of the exported clustering object

    Usage:
        exportClusteringObject(myObj, '/dir1', 'myObj')
    """

    if not os.path.exists(saveDir):
        os.makedirs(saveDir)

    fileName = saveDir + dataName + '_GroupsSubgroups.xlsx'

    writer = pd.ExcelWriter(fileName)

    for group in sorted([item for item in list(ClusteringObject.keys()) if not item=='linkage']):
        for subgroup in sorted([item for item in list(ClusteringObject[group].keys()) if not item=='linkage']):

            df_data = ClusteringObject[group][subgroup]['data'].iloc[::-1]
            df_dataAutocorr = ClusteringObject[group][subgroup]['dataAutocorr'].iloc[::-1]

            if includeData==True and includeAutocorr==True:
                df = pd.concat((df_data,df_dataAutocorr), sort = False, axis=1)
            elif includeData==True and includeAutocorr==False:
                df = df_data
            elif includeData==False and includeAutocorr==True:
                df = df_dataAutocorr
            else:
                df = pd.DataFrame(index=df_data.index)

            df.index.name = 'Index'
            df.to_excel(writer, 'G%sS%s'%(group, subgroup))

    writer.save()

    print('Saved clustering object to:', fileName)

    return fileName

def getCommunitiesOfTimeSeries(data, times, minNumberOfCommunities=2, horizontal=False, method='WDPVG', direction='left',weight='distance'):

    '''Get communities of time series

    Parameters:
        data: 1d numpy.array
            Data array

        times: 1d numpy.array
            Times corresponding to data points

        minNumberOfCommunities: int, Default 2
            Number of communities to find depends on the number of splits.
            This parameter is ignored in methods that automatically
            estimate optimal number of communities.
        
        horizontal: boolean, Default False
            Whether to use horizontal of normal visibility graph

        method: str, Default 'betweenness_centrality'
            Name of the method to use:
                'Girvan_Newman': edge betweenness centrality based approach

                'betweenness_centrality': reflected graph node betweenness centrality based approach

                'WDPVG': weighted dual perspective visibility graph method (also set weight variable)
    
        direction:str, default 'left'
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.
                'left': nodes can only aggregate to the left side hubs, e.g. early hubs
                'right': nodes can only aggregate to the right side hubs, e.g. later hubs
        
        weight: str, Default 'distance'
            Type of weight for method='WDPVG':
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5

    Returns:
        (list, graph)
            List of communities and a networkx graph

    Usage:
        res = getCommunitiesOfTimeSeries(data, times)
    '''

    if method=='betweenness_centrality':
        if horizontal:
            graph_nx = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfHVG(data))
            graph_nx_inv = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfHVG(-data))
        else:  
            graph_nx = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfNVG(data, times))
            graph_nx_inv = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfNVG(-data, times))
    
        def find_and_remove_node(graph_nx):
            bc = nx.betweenness_centrality(graph_nx)
            node_to_remove = list(bc.keys())[np.argmax(list(bc.values()))]
            graph_nx.remove_node(node_to_remove)
            return graph_nx, node_to_remove

        list_of_nodes = []
        for i in range(minNumberOfCommunities-1):
            graph_nx_inv, node = find_and_remove_node(graph_nx_inv)
            list_of_nodes.append(node)
        
        if not 0 in list_of_nodes:
            list_of_nodes.append(0)

        list_of_nodes.append(list(graph_nx.nodes)[-1] + 1)
        list_of_nodes.sort()

        communities = [list(range(list_of_nodes[i],list_of_nodes[i + 1])) for i in range(len(list_of_nodes) - 1)]

    elif method=='Girvan_Newman':
        if horizontal:
            graph_nx = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfHVG(data))
        else:
            graph_nx = nx.from_numpy_matrix(visibilityGraphAuxiliaryFunctions.getAdjacencyMatrixOfNVG(data, times))
        generator_of_communities = nx.algorithms.community.centrality.girvan_newman(graph_nx)

        for i in range(minNumberOfCommunities-1):
            communities_for_level = next(generator_of_communities)
        
        communities = list(sorted(c) for c in communities_for_level)

    elif method=='WDPVG':
        if horizontal:
            graph_nx = visibilityGraphCommunityDetection.createVisibilityGraph(data, times, "dual_horizontal", weight=weight)[0]
        else:
            graph_nx = visibilityGraphCommunityDetection.createVisibilityGraph(data, times, "dual_natural", weight=weight)[0]

        communities = visibilityGraphCommunityDetection.communityDetectByPathLength(graph_nx, direction=direction, cutoff='auto')

    else:

        print('Unknown method: %s'%(method))
        
        return None

    return communities, graph_nx


'''This module contains global constants used in PyIOmica.
Some of the modules, classes and functions are imported in this module.'''

import io
import os
import types
import appdirs
import gzip
import copy
import shutil
import zipfile

import numpy as np
np.random.seed(0)

import pandas as pd
import networkx as nx

import scipy
import scipy.stats
import scipy.fftpack
import scipy.cluster.hierarchy as hierarchy

import sklearn


printPackageGlobalDefaults = False
'''Whether to print package global defaults listed in this module'''

PackageDirectory = os.path.dirname(__file__)
'''Package directory'''

UserDataDirectory = os.path.join(PackageDirectory, 'data')

if not os.path.exists(UserDataDirectory):
    print('Cannot access default site package directory, will instead use ', UserDataDirectory)
    UserDataDirectory = appdirs.user_data_dir('pyiomica', 'gmiaslab')

ConstantPyIOmicaDataDirectory = UserDataDirectory
'''ConstantPyIOmicaDataDirectory is a global variable pointing to the PyIOmica data directory.'''

ConstantPyIOmicaExamplesDirectory = os.path.join(UserDataDirectory, "ExampleData")
'''ConstantPyIOmicaExamplesDirectory is a global variable pointing to the PyIOmica example data directory.'''

ConstantPyIOmicaExampleVideosDirectory = os.path.join(UserDataDirectory, "ExampleVideos")
'''ConstantPyIOmicaExampleVideosDirectory is a global variable pointing to the PyIOmica example videos directory.'''

for path in [ConstantPyIOmicaDataDirectory, ConstantPyIOmicaExamplesDirectory, ConstantPyIOmicaExampleVideosDirectory]:
    if not os.path.exists(path):
        os.makedirs(path)
 
del(path)

ConstantGeneDictionary = None
'''ConstantGeneDictionary is a global gene/protein dictionary variable typically created by GetGeneDictionary.'''



"""
setup.py for pyiomica package
"""
from setuptools import setup, find_packages
from codecs import open
from os import path
here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description=f.read()

setup(
    name='pyiomica',
    packages=find_packages(),
    version='1.3.1',
    description='Omics Analysis Tool Suite',
    long_description_content_type="text/markdown",
    long_description=long_description,
    include_package_data=True,
    author='S. Domanskyi, C. Piermarocchi, G. Mias',
    author_email='gmiaslab@gmail.com',
    license='MIT',
    url='https://github.com/gmiaslab/pyiomica',
    download_url='https://github.com/gmiaslab/pyiomica/archive/1.3.1.tar.gz',
    keywords=['omics', 'longitudinal','bioinformatics'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Topic :: Education',
        'Topic :: Utilities',
        ],
    install_requires=[
        'h5py>=2.9.0',
        'matplotlib>=3.1.0',
        'networkx>=2.3',
        'numba>=0.44.1',
        'numpy>=1.16.4',
        'openpyxl>=2.6.2',
        'pandas>=0.24.2',
        'pymysql>=0.9.3',
        'tables>=3.5.2',
        'requests>=2.22.0',
        'scikit-learn>=0.21.2',
        'scipy>=1.2.1',
        'xlsxwriter>=1.1.8',
        'appdirs>=1.4.3'],
    zip_safe=False
)



"""Visibility Graph Community detection functions.
"""
from .globalVariables import *
import numpy as np
import networkx as nx

def createVisibilityGraph(data, times, graph_type='natural', weight=None, withsign=False):

    """Calculate adjacency matrix of visibility graph, create the networkx.Graph network.

    Parameters:
        data: 2d numpy.array
            Data to process
        
        times: 1d numpy.array
            Times corresponding to provided data points
        
        weight: str, Default None
            Type of normalization:
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
            
        graph_type: str, Default 'natural'
            Type of the visibility graph:
                "horizontal", Horizontal Visibility Graph

                "natural", Natural Visibility Graph

                "dual_horizontal", Dual Perspective Horizontal Visibility Graph

                "dual_natural", Dual Perspective Natural Visibility Graph 
                
        withsign: boolean, Default False

            Whether to return the sign of adjacency matrix, 
            the link from normal perspective VG is positive,
            the link from reflected perspective VG is negative 


    Returns: tuple
        Tuple of two objects:
            networkx.Graph
                Graph of networkx type
            
            2d numpy.array
                Adjacency matrix

    Usage:
        A = createVisibilityGraph(data, times)
    """
    
    idx_nan = np.argwhere(np.isnan(data))
    ndata = np.delete(data, idx_nan)
    ntimes = np.delete(times, idx_nan)
    if graph_type not in ['horizontal','dual_horizontal','dual_natural','natural']:
        print('Unknown graph type: %s, adjust graph type to natural'%(graph_type))
    else:
        print('graph type is: %s' %(graph_type))
        
    if weight not in ['time', 'tan', 'distance', None]:
        print('Unknown weight type: %s, adjust weight to None'%(weight))
    else:
        print('weight is: %s' %(weight))
    
    if graph_type == "horizontal":
        AdMatrixOfVisibilityGraph = np.asmatrix(__getAdjacencyMatrixOfHorizontalVisibilityGraph(ndata, ntimes, weight=weight))
    elif graph_type == "dual_horizontal":
        AdMatrixOfVisibilityGraph = np.asmatrix(__getAdjacencyMatrixOfHorizontalVisibilityGraph_dual(ndata, ntimes, weight=weight, withsign=withsign))
    elif graph_type == "dual_natural":
        AdMatrixOfVisibilityGraph = np.asmatrix(__getAdjacencyMatrixOfVisibilityGraph_dual(ndata, ntimes, weight=weight, withsign=withsign))
    else:
        AdMatrixOfVisibilityGraph = np.asmatrix(__getAdjacencyMatrixOfVisibilityGraph(ndata, ntimes, weight=weight))
        
    G = nx.convert_matrix.from_numpy_matrix(abs(AdMatrixOfVisibilityGraph))
    
    labels = {}
    intensity = {} 
    for idx, node in enumerate(G.nodes()):
        labels[node] = str(np.round(ntimes[idx],2))
        intensity[node] = ndata[idx]
    nx.set_node_attributes(G, labels, 'timepoint')
    nx.set_node_attributes(G, intensity, 'intensity')
    
    
    return G,AdMatrixOfVisibilityGraph

def __getAdjacencyMatrixOfHorizontalVisibilityGraph(data, times, weight=None):

    """Calculate adjacency matrix of horizontal visibility graph.

    Parameters:
        data: 2d numpy.array
            Data to process
        
        times: 1d numpy.array
            Times corresponding to provided data points
        
        weight: str, Default None
            Type of normalization:
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = __getAdjacencyMatrixOfHorizontalVisibilityGraph(data, times)
    """

    dimension = len(data)

    A = np.zeros((dimension,dimension))

    for i in range(dimension):
        if i<dimension-1:
            if weight == 'time':
                A[i, i+1] = A[i+1, i] = abs(times[i] - times[i+1])
            elif weight == 'tan':
                A[i, i+1] = A[i+1, i] = abs((data[i] - data[i+1])/(times[i] - times[i+1])) + 10**(-8)
            elif weight == 'distance':
                A[i, i+1] = A[i+1, i] = ((data[i] - data[i+1])**2 + (times[i] - times[i+1])**2)**0.5
            else:
                A[i, i+1] = A[i+1, i] = 1


        for j in range(i + 2, dimension):
            if np.max(data[i+1:j]) < min(data[i], data[j]):
                if  weight == 'time':
                    A[i,j] = A[j,i] = abs(times[i] - times[j])
                elif weight == 'tan':
                    A[i, j] = A[j, i] = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)
                elif weight == 'distance':
                    A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
                else:
                    A[i, j] = A[j, i] = 1
    return A

def __getAdjacencyMatrixOfHorizontalVisibilityGraph_dual(data, times, weight=None, withsign=False):

    """Calculate adjacency matrix of dual perspective horizontal visibility graph.

    Parameters:
        data: 2d numpy.array
            Data to process
        
        times: 1d numpy.array
            Times corresponding to provided data points
        
        weight: str, Default None
            Type of normalization:
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
            
        withsign: boolean, Default False

            Whether to return the sign of adjacency matrix, the link from normal perspective VG is positive,
            the link from reflected perspective VG is negative        

        
    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = __getAdjacencyMatrixOfHorizontalVisibilityGraph_dual(data, times)
    """

    A_posi = np.asmatrix(__getAdjacencyMatrixOfHorizontalVisibilityGraph(data, times, weight=weight))
    A_nega = np.asmatrix(__getAdjacencyMatrixOfHorizontalVisibilityGraph(-data, times, weight=weight))
    A_dual = np.where(A_posi >= A_nega, A_posi, -A_nega)
    
    if withsign == True:
        A_dual = A_dual
    else:
        A_dual = abs(A_dual)
    
    return A_dual

def __getAdjacencyMatrixOfVisibilityGraph(data, times, weight=None):

    """Calculate adjacency matrix of natural visibility graph.

    Parameters:
        data: 2d numpy.array
            Data to process
        
        times: 1d numpy.array
            Times corresponding to provided data points
        
        weight: str, Default None
            Type of normalization:
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = __getAdjacencyMatrixOfVisibilityGraph(data, times)
    """

    dimension = len(data)

    V = (np.subtract.outer(data, data))/(np.subtract.outer(times, times) + np.identity(dimension))
    

    A = np.zeros((dimension,dimension))

    for i in range(dimension):
        if i<dimension-1:
            if weight == 'time':
                A[i, i+1] = A[i+1, i] = abs(times[i] - times[i+1])
            elif weight == 'tan':
                A[i, i+1] = A[i+1, i] = abs((data[i] - data[i+1])/(times[i] - times[i+1])) + 10**(-8)
            elif weight == 'distance':
                A[i, i+1] = A[i+1, i] = ((data[i] - data[i+1])**2 + (times[i] - times[i+1])**2)**0.5
            else:
                A[i, i+1] = A[i+1, i] = 1
                
        for j in range(i + 2, dimension):
            if np.max(V[i+1:j,i]) < V[j,i]:
                if  weight == 'time':
                    A[i,j] = A[j,i] = abs(times[i] - times[j])
                elif weight == 'tan':
                    A[i, j] = A[j, i] = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)
                elif weight == 'distance':
                    A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
                else:
                    A[i, j] = A[j, i] = 1

    return A

def __getAdjacencyMatrixOfVisibilityGraph_dual(data, times, weight=None, withsign=False):

    """Calculate adjacency matrix of dual perspective natural visibility graph.

    Parameters:
        data: 2d numpy.array
            Data to process
        
        times: 1d numpy.array
            Times corresponding to provided data points
        
        weight: str, Default None
            Type of normalization:
                None: no weighted

                'time': weight = abs(times[i] - times[j])

                'tan': weight = abs((data[i] - data[j])/(times[i] - times[j])) + 10**(-8)

                'distance': weight = A[i, j] = A[j, i] = ((data[i] - data[j])**2 + (times[i] - times[j])**2)**0.5
            
        withsign: boolean, Default False

            Whether to return the sign of adjacency matrix, the link from normal perspective VG is positive,
            the link from reflected perspective VG is negative 
            

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = __getAdjacencyMatrixOfVisibilityGraph_dual(data, times)
    """

    A_posi = np.asmatrix(__getAdjacencyMatrixOfVisibilityGraph(data, times, weight=weight))
    A_nega = np.asmatrix(__getAdjacencyMatrixOfVisibilityGraph(-data, times, weight=weight))
    if min(data)<0:
        A_dual = np.where(A_posi >= A_nega, A_posi, -A_nega)
    else:
        A_dual = A_posi
    if withsign == True:
        A_dual = A_dual
    else:
        A_dual = abs(A_dual)

    return A_dual

def communityDetectByPathLength(G, setSourceTarget=None,direction=None, cutoff=None):
    
    """Calculate community structure by shortest path length algorithm.
    
    Parameters:
        G: networkx.Graph
            Graph of networkx type
    
        direction:str, default None
            The direction that nodes aggregate to communities:
                None: no specific direction, e.g. both sides.
        
                'left': nodes can only aggregate to the left side hubs, e.g. early hubs
        
                'right': nodes can only aggregate to the right side hubs, e.g. later hubs
    
        cutoff: int, float or str, Default None
            Cutoff is used to combine initial communities, e.g. whenever the shortest path length of two adjacent hub nodes is smaller than cutoff, the communities with the two hub nodes will be combined:
        
                int or float: the percentile of all shortest path length distribution, between 0 ~ 100
        
                'auto': use optimized cutoff
        
                None: no cutoff
        
    Returns:
        list of list
            Detected communities in the form of nested list.

    Usage:
        c = communityDetectByPathLength(G)
    """

    if direction not in ['left', 'right', None]:
        print('Unknown direction type: %s, adjust direction to None'%(direction))
        direction = None
    else:
        print('direction type is: %s' %(direction))
        
        
    PL_dict = dict(nx.all_pairs_dijkstra_path_length(G)) # get shortest path length
    value_PL_list = []
    
    Nodes_PL_dict = list(PL_dict.keys())
    for node1 in range(len(PL_dict)):
        pl_node1 = PL_dict[Nodes_PL_dict[node1]]
        for node2 in range((node1+1), len(PL_dict)):
            value_PL_list.append(pl_node1[Nodes_PL_dict[node2]]) # get all path length value list
    
    nodelist = list(G.nodes)
    nodelist.sort()
    if setSourceTarget != None:
        PL_node0_node_end = nx.dijkstra_path(G, setSourceTarget[0], setSourceTarget[-1])
    else: 
        PL_node0_node_end = nx.dijkstra_path(G, nodelist[0], nodelist[-1]) #get path length from start to end time point
    
    community = {}

    sort_spl = PL_node0_node_end[:]
    sort_spl.sort()
    nodes_list = list(set(G.nodes) - set(PL_node0_node_end))

    print('the shortest path is:', sort_spl)
    
    #build community choose core as the nodes on the shortest path 
    #from start time point to end time point 
    for pl_node in sort_spl:
        community[pl_node] = [pl_node]
    
    #add nodes to community if the path length from nodes to core shorter than cutoff 
    for node in nodes_list:
        if direction == 'left':
            temp_spl = [i for i in sort_spl if i < node]
        elif direction == 'right':
            temp_spl = [i for i in sort_spl if i > node]
        else:
            temp_spl = sort_spl
            

        pl_to = {key:PL_dict[node][key] for key in temp_spl}
        pl_to_value = list(set(pl_to.values()))
        
        if 0 in pl_to_value:
            shortest_pl = min(pl_to_value.remove(0))
        else:
            shortest_pl = min(pl_to_value)
        c_key = []
        
        for k, v in pl_to.items():
            if v == shortest_pl:
                c_key.append(k)
        c_id = min(c_key)
        community[c_id].append(node)
    
    
    #merge communities if the path length between the cores shorter then cutoff
    if type(cutoff) == int or type(cutoff) == float:
        if 0.0 <= cutoff and cutoff <= 100.0:            
            cut = np.percentile(value_PL_list,cutoff)
            print('current percentiles cutoff is:', cutoff)
        else:
            cut = None
            print('Cutoff %f is out of range, adjust cutoff to None'%(cutoff))
    elif cutoff == 'auto':
        cut = __optimize_cutoff(PL_dict, value_PL_list, sort_spl)
        if max(value_PL_list) != min(value_PL_list):
            percentiles = (cut - min(value_PL_list)) / (max(value_PL_list) - min(value_PL_list)) * 100
        else:
            percentiles = 0
        print('current cutoff is auto, the optimized percentiles cutoff is %f ' %(percentiles))
    elif cutoff is None:
        cut = None
        print('current cutoff is None')
    else:
        cut = None
        print('Unknown cutoff type: %s, adjust cutoff to None'%(cutoff))

    if cut is not None:
            
        sort_spl_backup = sort_spl[:]
        
        while len(sort_spl) > 1:
            A = sort_spl[0]
            for B in sort_spl[1:]:
                if PL_dict[A][B] <= cut:
                    community[A].extend(community[B])
                    sort_spl_backup.remove(B)
                    del community[B]
            sort_spl_backup.remove(A)
            sort_spl = sort_spl_backup[:]
        
    c = list(community.values())
    for row in c:
        row.sort()

        
    return c

def __optimize_cutoff(PL_dict, value_PL_list, sort_spl):

    """Calculate the optimized cutoff, which will be used to combine initial communities. This function is used internally only.
    """

    L = len(sort_spl)
    plvalue = [min(value_PL_list)]
    for i in range (0,L-1):
        n1 = sort_spl[i]
        n2 = sort_spl[i+1]
        plvalue.append(PL_dict[n1][n2])
    plvalue.append(max(value_PL_list))
    plvalue.sort()
    plv_diff = [j-i for i, j in zip(plvalue[:-1], plvalue[1:])]
    
    index = plv_diff.index(max(plv_diff))
    plv_diff_copy = plv_diff[:]
    tcut = plvalue[index]

    while tcut > np.percentile(value_PL_list,50) and len(plv_diff)>1:
        plv_diff.pop()
        index = plv_diff_copy.index(max(plv_diff))
        tcut = plvalue[index]        


    if tcut <= np.percentile(value_PL_list,50):
        cut = tcut
    else:
        cut = None
    
    return cut



'''Utility functions'''


import multiprocessing

from .globalVariables import *


def readMathIOmicaData(fileName):

    '''Read text files exported by MathIOmica and convert to Python data

    Parameters:
        fileName: str
            Path of directories and name of the file containing data

    Returns:
        data
            Python data

    Usage:
        data = readMathIOmicaData("../../MathIOmica/MathIOmica/MathIOmicaData/ExampleData/rnaExample")
    '''

    if os.path.isfile(fileName):
        with open(fileName, 'r') as tempFile:
            data = tempFile.read()

        data = data.replace('\n','').replace('{','(').replace('}',')').replace('->',':').replace('|>','}')
        data = data.replace('<|','{').replace('^','*').replace('`','*').replace('Missing[]','"Missing[]"')
        data = data.replace("\\",'')
    else:
        print('File not found (%s)'%(fileName))

    returning = None

    try:
        returning = eval(data)
    except Exception as exception:
        print(exception)
        print('Error occured while converting data (%s)'%(fileName))

    return returning

def runCPUs(NumberOfAvailableCPUs, func, list_of_tuples_of_func_params):

    """Parallelize function call with multiprocessing.Pool.

    Parameters:
        NumberOfAvailableCPUs: int
            Number of processes to create

        func: function
            Function to apply, must take at most one argument

        list_of_tuples_of_func_params: list
            Function parameters

    Returns:
        2d numpy.array
            Results of func in a numpy array

    Usage:
        results = runCPUs(4, pAutocorrelation, [(times[i], data[i], allTimes) for i in range(10)])
    """

    instPool = multiprocessing.Pool(processes = NumberOfAvailableCPUs)
    return_values = instPool.map(func, list_of_tuples_of_func_params)
    instPool.close()
    instPool.join()

    return np.vstack(return_values)

def createReverseDictionary(inputDictionary):

    """Efficient way to create a reverse dictionary from a dictionary.
    Utilizes Pandas.Dataframe.groupby and Numpy arrays indexing.
    
    Parameters: 
        inputDictionary: dictionary
            Dictionary to reverse

    Returns:
        dictionary
            Reversed dictionary

    Usage:
        revDict = createReverseDictionary(Dict)
    """

    keys, values = np.array(list(inputDictionary.keys())), np.array(list(inputDictionary.values()))
    df = pd.DataFrame(np.array([[keys[i], value] for i in range(len(keys)) for value in values[i]]))
    dfGrouped = df.groupby(df.columns[1])
    keys, values = list(dfGrouped.indices.keys()), list(dfGrouped.indices.values())
    GOs = df.values.T[0]

    return dict(zip(keys, [GOs[value].tolist() for value in values]))

def createDirectories(path):

    """Create a path of directories, unless the path already exists.

    Parameters:
        path: str
            Path directory

    Returns:
        None

    Usage:
        createDirectories("/pathToFolder1/pathToSubFolder2")
    """

    if path=='':
        return None

    if not os.path.exists(path):
        os.makedirs(path)

    return None




'''Functions to generate adjacency matrix of visibility graphs'''

import numba
numba.config.NUMBA_DEFAULT_NUM_THREADS = 4

from .globalVariables import *

@numba.jit(cache=True)
def getAdjacencyMatrixOfNVG(data, times):

    """Calculate adjacency matrix of visibility graph.
    JIT-accelerated version (a bit faster than NumPy-accelerated version).
    Allows use of Multiple CPUs.

    Parameters:
        data: 2d numpy.array
            Numpy array of floats

        times: 1d numpy.array
            Numpy array of floats

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = getAdjacencyMatrixOfNVG(data, times)
    """

    dimension = len(data)

    V = np.zeros((dimension,dimension))

    for i in range(dimension):
        for j in range(i + 1, dimension):
            V[i,j] = V[j,i] = (data[i] - data[j]) / (times[i] - times[j])

    A = np.zeros((dimension,dimension))

    for i in range(dimension):
        for j in range(i + 1, dimension):
            no_conflict = True

            for a in list(range(i+1,j)):
                if V[a,i] > V[j,i]:
                    no_conflict = False
                    break

            if no_conflict:
                A[i,j] = A[j,i] = 1

    return A

def getAdjacencyMatrixOfNVGbyNUMPY(data, times):

    """Calculate adjacency matrix of visibility graph.
    NumPy-accelerated version. Somewhat slower than JIT-accelerated version.
    Use in serial applications.

    Parameters:
        data: 2d numpy.array
            Numpy array of floats

        times: 1d numpy.array
            Numpy array of floats

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = getAdjacencyMatrixOfNVGbyNUMPY(data, times)
    """

    dimension = len(data)

    V = (np.subtract.outer(data, data))/(np.subtract.outer(times, times) + np.identity(dimension))

    A = np.zeros((dimension,dimension))

    for i in range(dimension):
        if i<dimension-1:
            A[i,i+1] = A[i+1,i] = 1

        for j in range(i + 2, dimension):
            if np.max(V[i+1:j,i])<=V[j,i]:
                A[i,j] = A[j,i] = 1

    return A

@numba.jit(cache=True)
def getAdjacencyMatrixOfHVG(data):

    """Calculate adjacency matrix of horizontal visibility graph.
    JIT-accelerated version (a bit faster than NumPy-accelerated version).
    Single-threaded beats NumPy up to 2k data sizes.
    Allows use of Multiple CPUs.

    Parameters:
        data: 2d numpy.array
            Numpy array of floats

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = getAdjacencyMatrixOfHVG(data)
    """

    A = np.zeros((len(data),len(data)))

    for i in range(len(data)):
        for j in range(i + 1, len(data)):
            no_conflict = True

            for a in list(range(i+1,j)):
                if data[a] > data[i] or data[a] > data[j]:
                    no_conflict = False
                    break

            if no_conflict:
                A[i,j] = A[j,i] = 1

    return A

def getAdjacencyMatrixOfHVGbyNUMPY(data):

    """Calculate adjacency matrix of horizontal visibility graph.
    NumPy-accelerated version.
    Use with datasets larger than 2k.
    Use in serial applications.

    Parameters:
        data: 2d numpy.array
            Numpy array of floats

    Returns:
        2d numpy.array
            Adjacency matrix

    Usage:
        A = getAdjacencyMatrixOfHVGbyNUMPY(data)
    """

    dimension = len(data)

    A = np.zeros((dimension,dimension))

    for i in range(dimension):
        if i<dimension-1:
            A[i,i+1] = A[i+1,i] = 1

        for j in range(i + 2, dimension):
            if np.max(data[i+1:j])<=min(data[i], data[j]):
                A[i,j] = A[j,i] = 1

    return A


